// books2.cpp : An example of using lists as a container and maps
// as indices into the container.
// This is the second part and adds the maps
//

#include "stdafx.h"
#include <iostream>
#include <fstream>
#include <string>
#include <list>
#include <map>

using namespace std;
// this represents the generic object in the collection
// It is the parent class of the two classes that will actually be stored in the
// container. There will be no instances of this class.
class holding {
public:
	holding() {};
	holding(string n, int l, double c) { title = n; location = l; cost = c;}
	string get_title() { return title;}
	int get_location() { return location;};
	double get_cost() { return cost;}
	void set_title(string s) { title = s;}
	void set_location(int l) { location = l;}
	void set_cost( double c) { cost = c;}
	virtual void print()
	{
		cout << "Title: " << title;
		cout << " Cost: " << cost << " Location:" << location << endl;
	}

protected:
	string title;  // title
	int location;  // shelf number
	double cost;   // cost in dollars
}; // holding

class book : public holding {
public:
	book(string a, int p) {
		author = a; pages=p;
	}
	book(string a, int p, int loc, double c, string t) {
		author = a; pages=p;
		location = loc; cost = c; title = t;
	}
	int get_pages() { return pages;}
	string get_author() { return author;}
	void set_pages(int p) { pages = p;}
	void set_author(string a) { author = a;}
	void print() 
	{
		holding::print();  // run the parent version
		cout << " written by " << author;
		cout << " with " << pages << " pages." << endl;
		cout << "----------------------" << endl;
	} 

protected:
	string author;
	int pages;
}; // book


class video : public holding {
public:
	video(string a, int p) {
		director = a; length=p;
	}
	video(string a, int p, int loc, double c, string t) {
		director = a; length=p;
		location = loc; cost = c; title = t;
	}
	int get_length() { return length;}
	string get_director() { return director;}
	void set_length(int p) { length = p;}
	void set_director(string a) { director = a;}
	void print() 
	{
		holding::print(); // run the parent version
		cout << " Directed by " << director;
		cout << " length " << length << endl;
		cout << "----------------------" << endl;
	}
private:
	string director;
	int length;
}; // video

/*
 * This represents the collection as a whole.
 * We decided on two indices, one for books and one for
 * videos. One alternative was to create a single index
 * and tell the types of the objects wgen we access the
 * index. But this may involve using RTTI (Run Time Type
 * Identification) or require a type variable in the holding class.
 * We initially created the maps with a key of string and 
 * a value of either book or video pointer.
 * But later, when we wanted to do the adds we ran into a problem.
 * Since we use the same variable (h) of type holding * to store
 * the pointer returned by new, we would have had to cast it
 * to the type we needed (either book * or video *) to store
 * it in the map.
 * Instead we mape the two maps the same, storing a holding pointer
 * and we could just store h in the map.
*/
class library {
public:
	void load_data();
	void print_library();
	void get_book();
	void get_book2();
	void print_bookindex();
private:
	list<holding *> L;  // container of stuff
	map<string,holding *> bookindex;
	map<string,holding *> videoindex;
}; // library

void
library::load_data() 
{
	ifstream datain;
	char type;
	string author,director,title;
	int pages,location,length;
	double cost;
	holding *h;

	// open the data file.
	// There is one line per data entry
	// a book entry starts with a 'b' as the first entry and a video
	// starts with a 'v'.
	// Each then has the three holding data first in this order :
	//  title,location,cost
	//  one per line.
	//  These are then followed by the data that are unique to the 
	//  book and video classes.
	
	// first open the data file and check that it worked
	datain.open("books.txt");
	if(!datain) {
		cerr << "file open failed" << endl;
		exit(1);
	}

	// we loop forever so that we can check the eof status after the
	// first read. This prevents using the last record twice which happens if we use
	// while(!datain.eof()
	while(true) {
		datain >> type;
		if(datain.eof()) break;   // end of file?
		// now read the common data
		getline(datain,title,'$');
		title.erase(0,1);  // delete leading newline
		datain >> location;
		datain >> cost;

		if(type == 'b') {  // book record
			getline(datain,author,'$');
			author.erase(0,1);  // delete leading newline
			datain >> pages;
			h = new book(author,pages,location,cost,title);
			// The map object has an insert method to add a new entry.
			// The thing is adds is a value_type object. Like the iterator
			// class, this is defined inside the map class so we have to use the
			// :: operator to refer to it.
			// What we are doing here is essentially calling the constructor
			// for it and adding the new object to the map.
			bookindex.insert(map<string,holding *>::value_type(title,h));
		}
		else { // video
			getline(datain,director,'$');
			director.erase(0,1);  // delete leading newline
			datain >> length;
			h= new video(director,length,location,cost,title);
			videoindex.insert(map<string,holding *>::value_type(title,h));
		}
		L.push_back(h); // add the new object to the container
	} // read data
	datain.close();
} // load_data

// loop over the object in the container and print them.
// Note that the iterator is o holding iterator. Since the print
// function is virtual, we will call the print() method that is appropriate
// for the actual object the iterator points at, not the one that matches
// the type.
void
library::print_library()
{
	cout << "\n" << "List of titles " << endl;
		list<holding *>::iterator ptr = L.begin();
		while(ptr != L.end()) {
			(*ptr)->print();
			ptr++;
		}
} // print_library

// Prompt the user for a book title and use the index operator []
// to look up the book in the map.
// You get a NULL pointer back if it isn't found and
// a holding pointer if it is.
void
library::get_book()
{
	string s;
	holding * result;
	cout << "enter a book title ";
	cin >> s;
	result = bookindex[s];
	if(result == NULL) {
		cout << "no entry found" << endl;
		return;
	}
	result->print();
} // get_book

// test out the new bookindex map
void
library::get_book2()
{
	string s;
	cout << "enter a book title ";
	getline(cin,s);
	cout << "searching for :" << s << ":" << endl;
	// use the find method on the bookindex map.
	// It returns an iterator which points at the found
	// object or at end() if it wasn't found.
	map<string,holding *>::iterator it = bookindex.find(s);

    if(it == bookindex.end()) { // word not found
		cout << "no entry found" << endl;
		return;
	}
	// The thing in the map has two parts so it->second is
	// a pointer to the value part. In this case, this is the holding *.
	// So we have to chain two sets of ->.
	// The it->second part gets the holding * pointer.
	// Then we add another one to get to the print() method.
	it->second->print();
} //get_book2

// this uses an iterator to get each of the elements in the
// bookindex and print them in turn.
void
library::print_bookindex()
{
	map<string,holding*>::iterator it;
    for(it=bookindex.begin(); it != bookindex.end(); it++) {
        cout << ":" << it->first<< ":" << endl;
		it->second->print();
	}
} // print_bookindex

int 
main(int argc, char* argv[])
{
	library L;
	L.load_data();
	L.print_library();
	L.get_book2();
	L.print_bookindex();
	return 0;
} // main
