// fnobj.cpp CIS 275Function Objects

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <vector>
#include <algorithm>

using namespace std;

/*
In C, a function is a piece of executable code that
takes arguments and returns a value.
C++ abtracts this idea to include anything that implements
the () operator.
This allows us to create classes that include an overloaded
() operator can be used where a function can.
*/

// this class stores the high and low values of a range and 
// implements the () operator to check if something is in the range
template <class T>
class inrange {
   public:
      inrange( const T& L, const T& H) : low(L),high(H) {};
      bool operator()(const T& t) { 
	      return ((t > low) && ( t < high)) ; 
      }
   private:
      const T low;
      const T high;
};

// create a randome integer less than 10
int myrand() { 
   return(rand() % 10); 
} 

// this is the actual code for find_if
template <class InputIterator, class Predicate>
InputIterator find3(InputIterator first, InputIterator last,
                      Predicate pred) {
   while (first != last && !pred(*first)) ++first;
   return first;
}

int 
main(int argc, char* argv[])
{
   int size=10;  // default value
   if(argc > 1) size=atoi(argv[1]);
   vector<int> v(size);
   vector<int>::iterator iter;
   inrange<int> foo(1,5);

   // This looks like a function call to a piece of code called 'foo'.
   // But it is a call on the code for operator() defined in the class inrange.
   if(foo(3)) cout << "3 is in range" << endl;
   if(!foo(7)) cout << "7 is out of range" << endl;

   cout << "vector before generate is " << v.size() << endl;
   // fill the vector with random numbers
   // calls the myrand() function for each slot between
   // the begin and end iterators
   generate(v.begin(),v.end(),myrand);
   cout << "vector after generate is " << v.size() << endl;

   // print the vector first
   // This uses the third argument to copy elements from
   // the iterator range.
   // If the cout part was replaced with a filename, it would
   // write to the file.
   // The string arg is the separator
   copy(v.begin(),v.end(),ostream_iterator<int>(cout," "));
   cout << endl;

   // findif() applies the function object in the third
   // argument to eacj element in the iterator range.
   // If the function object returns true, it stop
   // looking and returns an iterator that points to the
   // found one.
    if(iter= find_if(v.begin(),v.end(),inrange<int>(0,5)))
       cout << "find_if: there is one in range " << *iter << endl;
    else
       cout << "find_if: there is nothing in range " << endl;

    if(iter= find3(v.begin(),v.end(),inrange<int>(0,5)))
       cout << "find3: there is one in range " << *iter << endl;
    else
       cout << "find3: there is nothing in range " << endl;

	return 0;
}
