// listhead - class that represents the list as a whole
#include "stdafx.h"
#include "listhead.h"
#include "data.h"

// have to destroy each one in the list
listhead::~listhead()
{
	node * ptr, *current;
	if(head == NULL)
		return;
	ptr = head;
	while(ptr != NULL) {
		current = ptr;
		ptr = ptr->get_next();
		delete current;
	} // while
} // listheader destructor

ostream& 
operator<<(ostream & os, listhead &lh)
{ 
   node * ptr=lh.head;
   int count = 0;
   os << "[";
   while(ptr != NULL) {
      if(count > 0) os << ",";
      os << ptr->get_data();
      count++;
      ptr = ptr->get_next();
   }
   os << "]";
   return os;
} // operator<<

// add a data element to the head of the list. Returns true if OK
bool
listhead::add_to_front(data nodedata)
{
	node * new_node;

	if( (new_node = new node) == NULL) {
		cerr << " no memory for node \n";
		return false;
	}
	new_node->set_data(nodedata);
	new_node->set_prev(NULL);
	new_node->set_next(head);
	if(head != NULL)
		head->set_prev(new_node);
	head = new_node;
	if(tail == NULL) tail = new_node;
	length += 1;
	return true;
} // add_to_front
//
// add a data element to the end of the list. Returns true if OK
bool
listhead::add_to_end(data nodedata)
{
	node * new_node;

	if( (new_node = new node) == NULL) {
		cerr << " no memory for node \n";
		return false;
	}
	new_node->set_data(nodedata);
	new_node->set_next(NULL);
	new_node->set_prev(tail);

	if(tail == NULL)
		head = tail = new_node;
	else
		tail->set_next(new_node);
	tail = new_node;
	length += 1;
	return true;
} // add_to_end

// search the list node by node until we find the one
// that contains my_d. Return NULL if we don't find it
node *
listhead::find_element(data my_d)
{
   node * lptr = head;
   while(lptr != NULL) {
      if( lptr->get_data() == my_d ){
         return(lptr); // found it
      }
      lptr = lptr->get_next();
   } // while
   return(NULL);
} // find_element

// search the list for the target data and add a new node after
// that one.
bool
listhead::add_after(data target,data my_d)
{
   node *lp,*lptr;
   
   if((lp = new node(my_d)) == NULL) return false;
   // use the find_element() function above to locate target
   // in the list, if it is there
   lptr = find_element(target);
   if(lptr == NULL)  return(false); // not there
      
   // the new one should point to the one
   // after the one we found (lptr->next)
   // and also to the target one
   lp->set_prev(lptr);
   lp->set_next(lptr->get_next());
   // the one after the target should point to the new one
   lptr->get_next()->set_prev(lp);
   // the one we found should point to the new one
   lptr->set_next(lp);
   // so the new one is after the one we found
   return true;   
} // add_after

// search the list for the target data and add a new node before
// that one.
bool
listhead::add_before(data target,data my_d)
{
   node *lp,*lptr;
   
   if((lp = new node(my_d)) == NULL) return false;
   // use the find_element() function above to locate target
   // in the list, if it is there
   lptr = find_element(target);
   if(lptr == NULL)  return(false); // not there

   // the new one should point to the one
   // before the one we found (lptr->next)
   // and also to the target one
   lp->set_prev(lptr->get_prev());
   lp->set_next(lptr);
   // the target should point to the new one
   lptr->set_prev(lp);
   // the previous one should point to the new one
   lptr->get_prev()->set_next(lp);
   // so the new one is before the one we found
   return true;   
} // add_before

// remove the node that contains the data my_d
bool
listhead::remove(data my_d)
{
   node *current;
   // if the first one is the one we want, delete it quick
   if( head->get_data() == my_d ){
      head = head->get_next();
      head->set_prev(NULL);
      return true;
   }

   // if we find it, make the one that comes before it
   // point to  the one that comes after. This deletes the node
   // from the list
   current = head;
   while(current != NULL) {
      if( current->get_data() == my_d ){ 
	 // make the one before the target point to the one after the
	 // target. And the one after should point to the one before
	 current->get_prev()->set_next(current->get_next() );
	 if(current != tail)  // last one
	    current->get_next()->set_prev(current->get_prev() );
	 delete current;
	 return true;
      }
      // move the two pointers to the next nodes
      current = current->get_next();
   } // while

   return false;  // didn't find it
} // remove
