// strcpy.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <stdio.h>
#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

const int STRLEN= 50;
ofstream ofile;  // output file

// gets enough space to hold a copy of the input string
char *
makespace(char * src)
{
   char * copy;
   if( (copy = new char[strlen(src) + 1]) == NULL) {
       cerr << "No memory for string:" << src << ":" << endl;
   }
   return copy;
} // makespace

// this copies using a for loop with the length
// calculated using strlen().
// By adding one, we copy the NUL at the end.
char *
forloop1(char * src)
{
   char *copy;
   if( (copy = makespace(src)) == NULL)  exit(1);
   for(int i=0; i < (int) strlen(src)+1; i++)
      copy[i] = src[i];
   return copy;
} // forloop1

// uses a for loop but the stop condition is
// when we see the NUL, not the length
char *
forloop2(char * src)
{
   char *copy;
   if( (copy = makespace(src)) == NULL)  exit(1);
   for(int i=0; src[i] != '\0'; i++)
      copy[i] = src[i];
   copy[i] = '\0'; // copy stops when it sees the NUL se we have to add it
   return copy;
} // forloop2

/* here is another way to do this. Suggested by a CIS327 student, Doug.
   it doesn't require the copy pointers since it doesn't alter
   the originals. Thus it uses somewhat less space.
   Plus there is no chance of accidentally altering the original
*/
char *
forloop3(char * src)
{
   char *copy;
   if( (copy = makespace(src)) == NULL)  exit(1);
   // note that there is no code in the body.
   // of the for loop, all the work is done in the stop condition
   // This works because assignment is an expression.
   for(  int x=0; (*(copy+x) = *(src+x)) != '\0'; x++) ;
   // we don't need to set the last NUL because the assignment
   // happens before the comparison, so the NUL gets copied.
   return copy;
} // forloop3

// just like forloop3(), but using array notation.
char *
forloop4(char * src)
{
   char *copy;
   if( (copy = makespace(src)) == NULL)  exit(1);
   for(  int y=0; (copy[y] = src[y]) != '\0'; y++) ;
   return copy;
} // forloop4

// pretty much like the for loops, but it uses a while
char *
whileloop(char * src)
{
   char *copy;
   if( (copy = makespace(src)) == NULL)  exit(1);
   int i =0;
   while(src[i]) {
      copy[i] = src[i];
      i++;
   }
   copy[i] = '\0'; // loop stop before NUL is copied
   return copy;
} // whileloop

// like whileloop(), but since the do loop checks at the
// bottom of the loop, we don't have to set the NUL
// Suggested by a CIS275 student.
char *
doloop(char * src)
{
   char *copy;
   if( (copy = makespace(src)) == NULL)  exit(1);
   int i =0;
   do {
      copy[i] = src[i];
   } while(src[i++]);
   return copy;
} // doloop

char *
hardest(char * src)
{
   char *copy,*dest;
   if( (copy = makespace(src)) == NULL)  exit(1);
   // we don't need a copy of src since it is already a copy
   dest=copy; 
   // first copy from src to dest, then increment both, then evaluate
   // the assignment for truth.
   // There is no code in the body of thw while loop.
   while(*dest++ = *src++) ; // all the work is in here
   return copy;
} // hardest

// like the one above, but with output
char *
hardest2(char * src)
{
   char *copy,*dest;
   if( (copy = makespace(src)) == NULL)  exit(1);
   ofile << "Dest start address is [" << (int)copy << "] ";
   ofile << "src start address is [" << (int)src << "]" << endl;
   dest=copy; 
   while(*dest++ = *src++) {
      ofile << "copied src[" << (int)src-1 << "](" << (char)*(src-1) << ")";
      ofile << " to dest[" << (int)dest-1 << "](" << (char)*(dest-1) << ")" << endl;
   }
   return copy;
} // hardest2

int 
main(int argc, char* argv[])
{
   char instr[STRLEN],*ptr;
   // setup the output file
   ofile.open("strcpy.txt");
   if(!ofile) {
      cerr << "File open failed" << endl;
      return(1);
   }

   (void) strcpy(instr,"forloop1");
   ofile << "the input string was :" << instr << ":" << endl;
   ofile << "copy using forloop1 :" << forloop1(instr) << ":" << endl;

   (void) strcpy(instr,"forloop2");
   ofile << "the input string was :" << instr << ":" << endl;
   ofile << "copy using forloop2 :" << forloop2(instr) << ":" << endl;

   (void) strcpy(instr,"forloop3");
   ofile << "the input string was :" << instr << ":" << endl;
   ofile << "copy using forloop3 :" << forloop3(instr) << ":" << endl;

   (void) strcpy(instr,"forloop4");
   ofile << "the input string was :" << instr << ":" << endl;
   ofile << "copy using forloop4 :" << forloop4(instr) << ":" << endl;

   (void) strcpy(instr,"whileloop");
   ofile << "the input string was :" << instr << ":" << endl;
   ofile << "copy using whileloop :" << whileloop(instr) << ":" << endl;

   (void) strcpy(instr,"doloop");
   ofile << "the input string was :" << instr << ":" << endl;
   ofile << "copy using doloop :" << doloop(instr) << ":" << endl;

   (void) strcpy(instr,"hardest");
   ofile << "the input string was :" << instr << ":" << endl;
   ofile << "copy using hardest :" << hardest(instr) << ":" << endl;

   // do it again, showing the steps
   (void) strcpy(instr,"hardest2");
   ofile << "the input string was :" << instr << ":" << endl;
   ofile << "copy using hardest2 :" << hardest2(instr) << ":" << endl;

   /* pointer loop printing */
   ofile << "pointer loop printing" << endl;
   (void) strcpy(instr,"print test");
   ptr=instr;
   while(*ptr) ofile << (*ptr++);
   ofile << '\n';

   return 0;
} // main

/*    HERE IS THE OUTPUT
the input string was :forloop1:
copy using forloop1 :forloop1:
the input string was :forloop2:
copy using forloop2 :forloop2:
the input string was :forloop3:
copy using forloop3 :forloop3:
the input string was :forloop4:
copy using forloop4 :forloop4:
the input string was :whileloop:
copy using whileloop :whileloop:
the input string was :doloop:
copy using doloop :doloop:
the input string was :hardest:
copy using hardest :hardest:
the input string was :hardest2:
Dest start address is [7930144] src start address is [6749636]
copied src[6749636](h) to dest[7930144](h)
copied src[6749637](a) to dest[7930145](a)
copied src[6749638](r) to dest[7930146](r)
copied src[6749639](d) to dest[7930147](d)
copied src[6749640](e) to dest[7930148](e)
copied src[6749641](s) to dest[7930149](s)
copied src[6749642](t) to dest[7930150](t)
copied src[6749643](2) to dest[7930151](2)
copy using hardest2 :hardest2:
pointer loop printing
print test
*/
