// tree.cpp - code for the tree class
#include "stdafx.h"
#include "tree.h"
#include "except.h"

bool
tree::add_value(char * str)
{
    if( (str == NULL) || (strlen(str) == 0) )
        throw new Err("add_value: bad parameter");
    return(add_valuer(root,str));
} // add_value

// this searches the tree for str and if it doesn't find
// it, it creates a node for it. It returns true if it
// either added a new node or found the value already in 
// the tree
bool
tree::add_valuer(node * &t,char * str)
{
    node * n;

    if(t == NULL) {
        if(debug) cout << "adding to NULL tree" << endl;
        if( (n = new node) == NULL) 
           throw new Mem("add_valuer: node");
        if( (n->value = new char[strlen(str)+1]) == NULL) 
           throw new Mem("add_valuer: string");
        (void) strcpy(n->value,str);
        n->left = n->right = NULL;
        t = n;
        return(true);
     }
     else {
        int c = strcmp(str,t->value);
        if(debug) cout << "root value is :" << t->value << ":\n";
         if(c == 0)  {
             if(debug) cout << "already in the tree" << endl;
             return(true);
         }
         else
              if ( c > 0) {
                 if(debug) cout << "adding right" << endl;
                 return(add_valuer(t->right,str));
              }
              else {
                 if(debug) cout << "adding left" << endl;
                 return(add_valuer(t->left,str));
              }
     }  // t not NULL
} // add_value

void
tree::print(enum PRINT_TYPE t)
{
    switch(t) {
        case INORDER:    // print inorder
	 	   cout << "INORDER Print" << endl;
                   break;
        case PREORDER:   // print preorder
	   	   cout << "PREORDER Print" << endl;
                   break;
        case POSTORDER:  // print postorder
	   	   cout << "POSTORDER Print" << endl;
                   break;
    } // esac
    printr(root,t);
} // print

void
tree::printr(node * r,enum PRINT_TYPE t)
{
    if(r == NULL) return;
    switch(t) {
        case INORDER:    // print inorder
                   printr(r->left,t);
                   cout << r->value << endl;
                   printr(r->right,t);
                   break;
        case PREORDER:   // print preorder
                   cout << r->value << endl;
                   printr(r->left,t);
                   printr(r->right,t);
                   break;
        case POSTORDER:  // print postorder
                   printr(r->left,t);
                   printr(r->right,t);
                   cout << r->value << endl;
                   break;
    } // esac
} // printr

void
tree::treeprint()
{
    treeprintr(root,0);
} // treeprint

// this prints the tree T sideways showing it's structure. 
// L is a helping variable to keep a count of the indentation to print
void
tree::treeprintr(node * t, int l)
{
   if(t != NULL) {
      treeprintr(t->right,l+1);
      for(int i=0; i< l; i++) cout << "  ";
      cout << t->value << endl;
      treeprintr(t->left,l+1);
   }
} // treeprintr

bool
tree::find(char * str)
{
    if( (str == NULL) || (strlen(str) == 0) )
       throw new Err("find: bad parameter");
    return(findr(root,str));
} // find

// this searches the tree for str 
// It returns true if it
// found the value and false otherwise
bool
tree::findr(node * t, char * str)
{

    if(t == NULL) return(false);
        
    int c = strcmp(str,t->value);
    if(c == 0)  return(true);  // found it
    if ( c > 0) 
        return(findr(t->right,str));
    else 
         return(findr(t->left,str));
} // findr

// Binary Tree Deletion

bool
tree::remove(char * val)
{
    if( (val == NULL) || (strlen(val) == 0) )
       throw new Err("remove: bad parameter");
    return(remover(val,root));
} // remove

void 
tree::del(node *& r, node * h)
{
   // r is passed by reference
   if(r == NULL) if(debug) cout << "in del, r is NULL" << endl;
   else
   	if(debug) cout << "in del: r->value is :" << r->value << ":" << endl;
   if(r->right != NULL)
       del(r->right,h);
   else {
       (void) strcpy(h->value,r->value);
       r = r->left;
   }
} // del

// T is passed by reference.
// Hunt through T for the node containing val and delete it.  
// Use a private method called del().
bool
tree::remover(char * val,node *& T)
{
    node * hold;
    if(T == NULL) return false;
           
    if(debug) cout << "t->value is :" << T->value << ":"  << endl;
    if(strcmp(val,T->value) < 0 ) {
       if(debug) cout << "deleting down the left" << endl;
       return(remover(val,T->left));
    }
    
    if(strcmp(val,T->value) > 0) {
       if(debug) cout << "deleting down the right" << endl;
       return(remover(val,T->right));
    }
    // found it
    if(debug) cout << "Found it" << endl;
    hold = T;   // remember where we are
    if(hold->right == NULL) {
       if(debug) cout << "Moving up the left" << endl;
       T = hold->left;   // move left child pointer up
       return true;
    }
    
    if(hold->left == NULL) {
      if(debug) cout << "Moving up the right" << endl;
      T = hold->right;   // move right child pointer up
      return true;
    }
     
    if(debug) cout << "Calling del()" << endl;
    del(hold->left,hold);
    return true;
} // remover

bool
tree::remove2(char * val)
{
    if( (val == NULL) || (strlen(val) == 0) )
       throw new Err("remove2: bad parameter");
    return(remover2(val,root));
} // remove2

void 
tree::del2(node *& r, node * h)
{
   // r is passed by reference
   if(r == NULL) if(debug) cout << "in del2, r is NULL" << endl;
   else
   	if(debug) cout << "in del2: r->value is :" << r->value << ":" << endl;
   if(r->left != NULL)
       del2(r->left,h);
   else {
       (void) strcpy(h->value,r->value);
       r = r->right;
   }
} // del2

// T is passed by reference.
// Hunt through T for the node containing val and delete it.  
// Use a private method called del().
bool
tree::remover2(char * val,node *& T)
{
    node * hold;
    if(T == NULL) return false;
           
    if(debug) cout << "t->value is :" << T->value << ":"  << endl;
    if(strcmp(val,T->value) < 0 ) {
       if(debug) cout << "deleting down the left" << endl;
       return(remover2(val,T->left));
    }
    
    if(strcmp(val,T->value) > 0) {
       if(debug) cout << "deleting down the right" << endl;
       return(remover2(val,T->right));
    }
    // found it
    if(debug) cout << "Found it" << endl;
    hold = T;   // remember where we are
    if(hold->left == NULL) {
       if(debug) cout << "Moving up the left" << endl;
       T = hold->right;   // move left child pointer up
       return true;
    }
    
    if(hold->right == NULL) {
      if(debug) cout << "Moving up the right" << endl;
      T = hold->left;   // move right child pointer up
      return true;
    }
     
    if(debug) cout << "Calling del()" << endl;
    del2(hold->right,hold);
    return true;
} // remover

/*
This is an iterative version of the binary tree insertion algorithm.
This is most closely related to the search-insert routine in the 
recursive tree routines handout, Trees 1.
This particular implementation is from "Applied Data Structures Using Pascal",
Guy J. Hale, Richard J. Easton, D. C. Heath and Company,1987.
*/
void
tree::inserti(char * val)
{
	insert2(root,val);
} // inserti

void
tree::insert2(node * &root,char *val)
{
        node * newnode, *top;
        bool inserted = false;

        if( (newnode = new node) == NULL) 
           throw new Mem("inserti: node");
        if( (newnode->value = new char[strlen(val)+1]) == NULL) 
           throw new Mem("inserti: string");
        (void) strcpy(newnode->value,val);
        newnode->right = newnode->left = NULL;
        top = root;
        if(root == NULL)
            root = newnode;
        else {
            while(!inserted) {
                if(strcmp(val,top->value) < 0)
                    if(top->left != NULL)
                        top = top->left;
                    else {
                        top->left = newnode;
                        inserted = true;
		    }
                else
                    if(top->right != NULL)
                        top = top->right;
                    else {
                        top->right = newnode;
                        inserted = true;
		    }
	    } // while
	} // root not NULL
} // insert2
