// tree.cpp - code for the tree class
#include "stdafx.h"
#include "tree.h"
#include "except.h"

// this searches the tree for str and if it doesn't find
// it, it creates a tree for it. It returns true if it
// either added a new tree or found the value already in 
// the tree
bool
add_value(tree * &t,char * str)
{
    tree * n;

    if( (str == NULL) || (strlen(str) == 0) )
        throw new Err("add_value: bad parameter");
    if(t == NULL) {
        if(Debug) cout << "adding to NULL tree" << endl;
        if( (n = new tree) == NULL) 
           throw new Mem("add_valuer: tree");
        if( (n->value = new char[strlen(str)+1]) == NULL) 
           throw new Mem("add_valuer: string");
        (void) strcpy(n->value,str);
        n->left = n->right = NULL;
        t = n;
        return(true);
     }
     else {
        int c = strcmp(str,t->value);
        if(Debug) cout << "root value is :" << t->value << ":\n";
         if(c == 0)  {
             if(Debug) cout << "already in the tree" << endl;
             return(true);
         }
         else
              if ( c > 0) {
                 if(Debug) cout << "adding right" << endl;
                 return(add_value(t->right,str));
              }
              else {
                 if(Debug) cout << "adding left" << endl;
                 return(add_value(t->left,str));
              }
     }  // t not NULL
} // add_value

void
inprint(tree * r,ofstream out)
{
     if(r == NULL) return;
     inprint(r->left,out);
     out << r->value << endl;
     inprint(r->right,out);
} // inprint

void
preprint(tree * r,ofstream out)
{
     if(r == NULL) return;
     out << r->value << endl;
     preprint(r->left,out);
     preprint(r->right,out);
} // preprint

void
postprint(tree * r,ofstream out)
{
     if(r == NULL) return;
     postprint(r->left,out);
     postprint(r->right,out);
     out << r->value << endl;
} // postprint

void
treeprint(tree * t,ofstream out)
{
	treeprintr(t,0,out);
}

// this prints the tree T sideways showing it's structure. 
// L is a helping variable to keep a count of the indentation to print
void
treeprintr(tree * t, int l,ofstream out)
{
   if(t != NULL) {
      treeprintr(t->right,l+1,out);
      for(int i=0; i< l; i++) out << "  ";
      out << t->value << endl;
      treeprintr(t->left,l+1,out);
   }
} // treeprintr

// this searches the tree for str 
// It returns true if it
// found the value and false otherwise
bool
find(tree * t, char * str)
{

    if(t == NULL) return(false);
    if( (str == NULL) || (strlen(str) == 0) )
       throw new Err("find: bad parameter");
        
    int c = strcmp(str,t->value);
    if(c == 0)  return(true);  // found it
    if ( c > 0) 
        return(find(t->right,str));
    else 
         return(find(t->left,str));
} // find

// Binary Tree Deletion

void 
del(tree *& r, tree * h)
{
   // r is passed by reference
   if(r == NULL) if(Debug) cout << "in del, r is NULL" << endl;
   else
   	if(Debug) cout << "in del: r->value is :" << r->value << ":" << endl;
   if(r->right != NULL)
       del(r->right,h);
   else {
       (void) strcpy(h->value,r->value);
       r = r->left;
   }
} // del

// T is passed by reference.
// Hunt through T for the tree containing val and delete it.  
// Use a private method called del().
bool
remove(char * val,tree *& T)
{
    tree * hold;
    if(T == NULL) return false;
    if( (val == NULL) || (strlen(val) == 0) )
       throw new Err("remove: bad parameter");
           
    if(Debug) cout << "t->value is :" << T->value << ":"  << endl;
    if(strcmp(val,T->value) < 0 ) {
       if(Debug) cout << "deleting down the left" << endl;
       return(remove(val,T->left));
    }
    
    if(strcmp(val,T->value) > 0) {
       if(Debug) cout << "deleting down the right" << endl;
       return(remove(val,T->right));
    }
    // found it
    if(Debug) cout << "Found it" << endl;
    hold = T;   // remember where we are
    if(hold->right == NULL) {
       if(Debug) cout << "Moving up the left" << endl;
       T = hold->left;   // move left child pointer up
       return true;
    }
    
    if(hold->left == NULL) {
      if(Debug) cout << "Moving up the right" << endl;
      T = hold->right;   // move right child pointer up
      return true;
    }
     
    if(Debug) cout << "Calling del()" << endl;
    del(hold->left,hold);
    return true;
} // remove

/*
This is an iterative version of the binary tree insertion algorithm.
This is most closely related to the search-insert routine in the 
recursive tree routines handout, Trees 1.
This particular implementation is from "Applied Data Structures Using Pascal",
Guy J. Hale, Richard J. Easton, D. C. Heath and Company,1987.
*/

void
inserti(tree * &root,char *val)
{
        tree * newtree, *top;
        bool inserted = false;

        if( (newtree = new tree) == NULL) 
           throw new Mem("inserti: tree");
        if( (newtree->value = new char[strlen(val)+1]) == NULL) 
           throw new Mem("inserti: string");
        (void) strcpy(newtree->value,val);
        newtree->right = newtree->left = NULL;
        top = root;
        if(root == NULL)
            root = newtree;
        else {
            while(!inserted) {
                if(strcmp(val,top->value) < 0)
                    if(top->left != NULL)
                        top = top->left;
                    else {
                        top->left = newtree;
                        inserted = true;
		    }
                else
                    if(top->right != NULL)
                        top = top->right;
                    else {
                        top->right = newtree;
                        inserted = true;
		    }
	    } // while
	} // root not NULL
} // inserti
